/* $Header: /cvsroot/sid/sid/user/c_tsw.c,v 1.17 2004/02/22 13:07:32 deppeler Exp $ */

#include "c_tsw.h"
#include "sha1_kernel.h"

/* globals - are these evil by nature? */
config_t	     *config; 
state_t		     *state;
int debug = 0;

int main (int argc, char **argv) {
	struct file_spec 	s_file;
	struct stat 		stats;
	struct cmd_line_t 	*cmd_line_opts;
	extern FILE 		*yyin;
	extern int		yyparse();

	config = (config_t *) ecalloc(1, sizeof(config_t));

	setprogname("tsw");

	/* process command line */
	cmd_line_opts = cmd_parse(argc, argv);
	if ( cmd_line_opts->debug != NULL )
		debug = cmd_line_opts->debug;


	/* read in configuration data */
	if ( cmd_line_opts->file != NULL ) {
		yyin = fopen(cmd_line_opts->file, "r");
	} else {
		printf("error: no configuration file specified\n");
		usage();
		exit(-1);
	}
	if ( yyin == NULL ) {
		eprintf("error: could not open file %s", cmd_line_opts->file);
	}
	yyparse();
	fclose(yyin);

	preprocess_time_spec(config);
	preprocess_syslog_spec(config);
	maint_pass(config);

	/* fork to background */
	if ( cmd_line_opts->debug == 0 ) {
		switch(fork()) {
			case -1:
				return (-1);
			case 0:
				printf("%s started.\n", progname());
				break;
			default:
				exit(0);
		}
	}

	/* tail on file */
	open_log(&s_file, &stats);
	tail(s_file, stats);
	close (s_file.fd);
	
	free_all(config);
	return 0;
}

/* 
 * tail on the file forever 
 */
int tail(struct file_spec s_file, struct stat stats) {

	lseek(s_file.fd, 0, SEEK_END);
	while ( 1 ) {

        	if (fstat (s_file.fd, &stats) < 0) {
        		s_file.fd = -1; 
			perror("cannot get stats");
			exit(2);
		}

		/* file truncated */
		if (stats.st_size < s_file.size) {
			dprintf(1, "file was truncated");
			fflush(stdout);
			lseek(s_file.fd, stats.st_size, SEEK_SET);
			s_file.size = stats.st_size;
		}

		if (stats.st_size == s_file.size) {
			dprintf(2, "sleeping");
			fflush(stdout);
			sleep(1);
			dprintf(2, "slept");
		} else {
			dprintf(2, "dumping");
			fflush(stdout);
		        s_file.size += dump_remainder(s_file.fd, stats.st_size - s_file.size);
		}
	

	}
	return 1;
}

/* read and process n_bytes of file starting at the current
 * position in fd
 * return the number of bytes read from the file.  
 */
long dump_remainder (int fd, off_t n_bytes)
{
	char buffer[BUFSIZ];
	int bytes_read;
	long n_written;
	off_t n_remaining = n_bytes;

	n_written = 0;
	while (1) {
		long n = MIN (n_remaining, (off_t) BUFSIZ);
		bytes_read = read (fd, buffer, n);
		if (bytes_read <= 0)
			break;

		if ( proc_buf(buffer, bytes_read) != 0 ) 
			printf("error processing buffer %s\n", buffer);

		dprintf(8, "bytes read: %d", bytes_read);
		n_remaining -= bytes_read;
		dprintf(8, "n_remaining: %d", n_remaining);
		n_written += bytes_read;
	}

	return n_written;
}


/* sort out individual lines from the buffer reads on the file
   process the lines as we encounter them */
int proc_buf(char *buf, int bytes_read) {
	int i = 0;
	int j = 0;
	char line[BUFSIZ];
	buf[bytes_read] = '\0';
	dprintf(8, "processing buffer:\n%s", buf);
	while ( i < bytes_read ) {
		line[j++] = buf[i];
		if ( buf[i] == '\n' || i == bytes_read - 1 ) {
			line[j-1] = '\0';
			dprintf(8, "got a line: %s", line);
			proc_line(line, j);
			j = 0;
		}
		i++;
	}
	return 0;
}

/* sort out the lines we're interested in
   open new pty: add to state table
   hash line: check validity
   close pty: remove from state table */
int proc_line(char *line, int length) {
	
	if (length < MIN_DIGEST_LINE_LENGTH) {
		dprintf(7, "proc_line: short line");
		return 0;
	}

	if ( pty_hash(line, length) ) 
		dprintf(7, "proc_line: pty hash");

	return 1;
}

/* check whether this is a currently expected hash 
   - get the state of this (host, pty) from the state table
   - check whether we're in maintenance or going into it now
   else check whether this is a valid command */
int check_digest_validity(char *hash, char *host, int pty) {
	state_t *s = NULL;

	dprintf(6, "checking digest validity");

	s = get_state(state, host, pty);

	if ( s == NULL ) {
		dprintf(7, "corresponding state not found");
		/* this happens when the parsing program is started
		after a pty has been opened and has never been
		allocated a state */
		s = add_state(state, host, pty);
		/* TODO: what to do when not in list: log something? */
	}
	dprintf(7, "corresponding state found");

	if ( check_maintenance(config, s, hash) ) {
		return 1;
	} else if ( check_hash(config, s, hash) ) {
		return 1;
	} 
	
	dprintf(7, "alarm!");

	return 0;
}


/* extract *format from *line into *tmp
   return > 0 on successful match
   return 0 on matching failure */
int find_entity(char *line, int length, char *host, const char *position, const char *format, char *tmp) {
	char *tline;

	/* find hostname */
	if ( sscanf(line, "%*s %*d %*d:%*d:%*d %s tsw: [", host) == 0) {
		printf("no host found in line!\n");
		line[0] = '\0';
		host[0] = '\0';
		tmp = '\0';
		return 0;
	}

	/* position tline */
	/* FIXME: we should scan without positioning first */
	tline =	strstr(line, position);
	if (tline != NULL) {
		dprintf(8, "find_entity: format: '%s'", format);

		if( sscanf(tline, format, tmp) ) {
			dprintf(9, "find_entity: tmp: '%s'", tmp);
			return 1;
		}
	}

	dprintf(8, "find_entity: no match!");
	tmp = '\0';
	host[0] = '\0';
	return 0;
}
	
/* search for a hash and a tty
   check validity if found */
int pty_hash(char *line, int length) {
	char host[HASH_LENGTH];
	char hash[HASH_LENGTH];
	char tmp[MAX_LINE_LENGTH];
	char tmp1[MAX_LINE_LENGTH];
	int  tty;
	long lines;
	host_cfg_t *hostp;

	dprintf(8, "pty_hash");
	if ( find_entity(line, length, host, MATCH_HASH, FMT_HASH, hash) && 
			find_entity(line, length, host, MATCH_HASH_TTY, FMT_HASH_TTY, tmp) &&
			find_entity(line, length, host, MATCH_HASH_LINE, FMT_HASH_LINE, tmp1)
	   ) {
		hash[44] = '\0'; /* add terminating null as hash scan was done with %c */
		dprintf(8, "pty_hash: hash: %s", hash);
		dprintf(8, "pty_hash: host: %s", host);
		tty = atoi(tmp);
		dprintf(8, "pty_hash: tty: %d", tty);
		lines = atol(tmp1);
		dprintf(8, "pty_hash: line: %ld", lines);

		/* 
		 * allocate state if necessary - state might not be allocated
		 * yet (e.g. session exists before this runs 
		 */
		if ( state == NULL ) {
			dprintf(9, "pty_hash: state not yet allocated. Allocating one ...");
			state = add_state(state, host, tty);
		}

		if ( check_digest_validity(hash, host, tty) ) {
			dprintf(1, "pty_hash: hash ok");
			log_ok(get_host(config->host, host));
		} else {
			dprintf(1, "pty_hash: hash not ok");
			hostp = get_host(config->host, host);
			log_alert(get_host(config->host, host));
			/* execute automated action if number of allowed attempts is exceeded */
			if ( state->failures >= hostp->max_failures )
				automated_action(get_state(state, host, tty), hostp);
		}
		return 1;
	} else {
		dprintf(8, "pty_open: find_entity failed!");
	}
	return 0;
}

/* open file, store fd, and stat and store its size */
int open_log(struct file_spec *s_file, struct stat *stats) 
{
	s_file->fd = open(config->logfile, O_RDONLY);
	if ( s_file->fd < 0 ) {
		perror("open_log: cannot open file");
		exit(1);
	}
        if (fstat (s_file->fd, stats) < 0) {
        	s_file->fd = -1; 
		perror("cannot get stats");
		exit(2);
	}
	s_file->size = stats->st_size;

	return 0;
}
